(** OCaml bindings for the GLX library. *)
(* Copyright (C) 2008 by Florent Monnier <fmonnier@linux-nantes.org>
 *
 * OCaml-Xlib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * OCaml-Xlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with OCaml-Xlib.  If not, see:
 * <http://www.gnu.org/licenses/>
 *)

(**
{ul
  {- {{:http://www.opengl.org/documentation/specs/glx/glx1.4.pdf}GLX Specs (pdf)}}
  {- {{:http://www.opengl.org/sdk/docs/man/xhtml/#glX}GLX Mans (on opengl.org)}}
  {- {{:http://publib.boulder.ibm.com/infocenter/systems/index.jsp?topic=/com.ibm.aix.opengl/doc/openglrf/OpenGLXSubs.htm}IBM Reference}}
}
*)
(*
  http://techpubs.sgi.com/library/tpl/cgi-bin/getdoc.cgi/0650/bks/SGI_Developer/books/OpenGL_Porting/sgi_html/ch04.html
*)

type glXContext  (* pointer to a structure *)

external glXQueryExtension: dpy:Xlib.display -> unit = "ml_glXQueryExtension"


external glXQueryVersion: dpy:Xlib.display -> int * int = "ml_glXQueryVersion"
(** returns (major, minor) *)

type client_name =
  | GLX_VENDOR
  | GLX_VERSION
  | GLX_EXTENSIONS
external glXGetClientString: dpy:Xlib.display -> name:client_name -> string = "ml_glXGetClientString"


(** {4 glXDrawable} *)

type glXDrawable
let glXDrawable_of_window ~(win : Xlib.window) = (Obj.magic win : glXDrawable) ;;

type glXWindow

external glXSwapBuffers: dpy:Xlib.display -> drawable:glXDrawable -> unit = "ml_glXSwapBuffers"

external glXWaitGL: unit -> unit = "ml_glXWaitGL"
external glXWaitX: unit -> unit = "ml_glXWaitX"

external glXDestroyWindow: dpy:Xlib.display -> win:glXWindow  -> unit = "ml_glXDestroyWindow"

type uint = int

type visual_attrib =
  | GLX_USE_GL
  | GLX_BUFFER_SIZE of uint
  | GLX_LEVEL of int
  | GLX_RGBA
  | GLX_DOUBLEBUFFER
  | GLX_STEREO
  | GLX_AUX_BUFFERS of uint
  | GLX_RED_SIZE of uint
  | GLX_GREEN_SIZE of uint
  | GLX_BLUE_SIZE of uint
  | GLX_ALPHA_SIZE of uint
  | GLX_DEPTH_SIZE of uint
  | GLX_STENCIL_SIZE of uint
  | GLX_ACCUM_RED_SIZE of uint
  | GLX_ACCUM_GREEN_SIZE of uint
  | GLX_ACCUM_BLUE_SIZE of uint
  | GLX_ACCUM_ALPHA_SIZE of int

external glXChooseVisual: dpy:Xlib.display -> screen:Xlib.screen_number -> attrib:visual_attrib list -> Xlib.xVisualInfo = "ml_glXChooseVisual"

type glXFBConfig
external glXCreateWindow: dpy:Xlib.display -> config:glXFBConfig -> win:Xlib.window  -> glXWindow = "ml_glXCreateWindow"

external glXMakeCurrent: dpy:Xlib.display -> drawable:glXDrawable -> ctx:glXContext -> unit = "ml_glXMakeCurrent"

external glXIsDirect: dpy:Xlib.display -> ctx:glXContext -> bool = "ml_glXIsDirect"

external glXCreateContext: dpy:Xlib.display -> vis:Xlib.xVisualInfo -> share_list:glXContext option -> direct:bool -> glXContext = "ml_glXCreateContext"

external glXCopyContext: dpy:Xlib.display -> ctx:glXContext -> mask:int -> glXContext = "ml_glXCopyContext"

external glXDestroyContext: dpy:Xlib.display -> ctx:glXContext -> unit = "ml_glXDestroyContext"

(** {3 Pbuffer} *)

type pbuf_attrib =
  | GLX_PBUFFER_WIDTH of int
  | GLX_PBUFFER_HEIGHT of int
  | GLX_LARGEST_PBUFFER of bool
  | GLX_PRESERVED_CONTENTS of bool

type glXPbuffer
external glXCreatePbuffer: dpy:Xlib.display -> config:glXFBConfig -> glXPbuffer = "ml_glXCreatePbuffer"
external glXDestroyPbuffer: dpy:Xlib.display -> pbuf:glXPbuffer -> unit = "ml_glXDestroyPbuffer"


