(** OCaml bindings for the Xlib library. *)
(*  Copyright (C) 2008 by Florent Monnier <fmonnier@linux-nantes.org>
 *
 * OCaml-Xlib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * OCaml-Xlib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with OCaml-Xlib.  If not, see:
 * <http://www.gnu.org/licenses/>
 *)

(** {3 Types} *)

type event_mask =
  | KeyPressMask
  | KeyReleaseMask
  | ButtonPressMask
  | ButtonReleaseMask
  | EnterWindowMask
  | LeaveWindowMask
  | PointerMotionMask
  | PointerMotionHintMask
  | Button1MotionMask
  | Button2MotionMask
  | Button3MotionMask
  | Button4MotionMask
  | Button5MotionMask
  | ButtonMotionMask
  | KeymapStateMask
  | ExposureMask
  | VisibilityChangeMask
  | StructureNotifyMask
  | ResizeRedirectMask
  | SubstructureNotifyMask
  | SubstructureRedirectMask
  | FocusChangeMask
  | PropertyChangeMask
  | ColormapChangeMask
  | OwnerGrabButtonMask

type uint = int
type window
type gc
type colormap

(** {3 Display} *)

type display
external xOpenDisplay: name:string -> display = "ml_XOpenDisplay"
external xCloseDisplay: dpy:display -> unit = "ml_XCloseDisplay"
external xFlush: dpy:display -> unit = "ml_XFlush"

external xProtocolVersion: dpy:display -> int = "ml_XProtocolVersion"
external xProtocolRevision: dpy:display -> int = "ml_XProtocolRevision"
external xVendorRelease: dpy:display -> int = "ml_XVendorRelease"
external xServerVendor: dpy:display -> string = "ml_XServerVendor"

external xlibSpecificationRelease: unit -> int = "ml_XlibSpecificationRelease"


(** {3 Screen number} *)

type screen_number
external xDefaultScreen: dpy:display -> screen_number = "ml_XDefaultScreen"

external xDisplayWidth: dpy:display -> screen_number -> int = "ml_XDisplayWidth"
external xDisplayHeight: dpy:display -> screen_number -> int = "ml_XDisplayHeight"

external xDefaultDepth: dpy:display -> screen_number -> int = "ml_XDefaultDepth"


(** {3 Pixel Colors} *)

type pixel_color
external xBlackPixel: dpy:display -> screen_number -> pixel_color = "ml_XBlackPixel"
external xWhitePixel: dpy:display -> screen_number -> pixel_color = "ml_XWhitePixel"


(** {3 XColor} *)

type xColor
external new_xColor: unit -> xColor = "ml_alloc_XColor"
(** this type is garbage collected *)

external xColor_set_red: xColor -> int -> unit = "ml_xColor_set_red"
external xColor_set_green: xColor -> int -> unit = "ml_xColor_set_green"
external xColor_set_blue: xColor -> int -> unit = "ml_xColor_set_blue"

type color_flags =
  | DoRed
  | DoGreen
  | DoBlue

external xColor_set_flags: xColor -> color_flags list -> unit = "ml_xColor_set_flags"

external xAllocColor: dpy:display -> colormap:colormap -> xColor -> unit = "ml_xAllocColor"

external xColor_pixel: xColor -> pixel_color = "ml_xColor_pixel"


(** {3 Visual} *)

type visual
external xDefaultVisual: dpy:display -> screen_number -> visual = "ml_XDefaultVisual"


(** {3 xVisualInfo} *)

type visualID

type xVisualInfo  (* pointer to a structure *)
type xVisualInfo_contents = { (* contents of this structure *)
    visual: visual;
    visual_id: visualID;
    screen_number: screen_number;
    depth: int;
    red_mask: uint;
    green_mask: uint;
    blue_mask: uint;
    colormap_size: int;
    bits_per_rgb: int;
  }

external xVisualInfo_datas: visual_info:xVisualInfo  -> xVisualInfo_contents = "ml_XVisualInfo_contents"
external xFree_xVisualInfo: xVisualInfo -> unit = "ml_XFree_XVisualInfo"


(** {3 Colormap} *)

type colormap_alloc =
  | AllocNone
  | AllocAll

external xCreateColormap: dpy:display -> win:window -> visual:visual -> alloc:colormap_alloc -> colormap = "ml_XCreateColormap"
external xFreeColormap: dpy:display -> colormap:colormap -> unit = "ml_XFreeColormap"
external xSetWindowColormap: dpy:display -> win:window -> colormap:colormap -> unit = "ml_XSetWindowColormap"


(** {4 xSetWindowAttributes} *)

type xSetWindowAttributes
external new_xSetWindowAttributes: unit -> xSetWindowAttributes = "ml_XSetWindowAttributes_alloc"
(** this type is garbage collected *)


type event_mask_list = event_mask list

type pixmap
type cursor

(* {{{ setting xSetWindowAttributes fields *)

external winAttr_set_background_pixmap: xSetWindowAttributes -> background_pixmap:pixmap -> unit = "ml_xSetWindowAttributes_set_background_pixmap"
external winAttr_set_background_pixel: xSetWindowAttributes -> background_pixel:pixel_color -> unit = "ml_xSetWindowAttributes_set_background_pixel"
external winAttr_set_border_pixmap: xSetWindowAttributes -> border_pixmap:pixmap -> unit = "ml_xSetWindowAttributes_set_border_pixmap"
external winAttr_set_border_pixel: xSetWindowAttributes -> border_pixel:pixel_color -> unit = "ml_xSetWindowAttributes_set_border_pixel"
external winAttr_set_bit_gravity: xSetWindowAttributes -> bit_gravity:int -> unit = "ml_xSetWindowAttributes_set_bit_gravity"
external winAttr_set_win_gravity: xSetWindowAttributes -> win_gravity:int -> unit = "ml_xSetWindowAttributes_set_win_gravity"
external winAttr_set_backing_store: xSetWindowAttributes -> backing_store:int -> unit = "ml_xSetWindowAttributes_set_backing_store"
external winAttr_set_backing_planes: xSetWindowAttributes -> backing_planes:uint -> unit = "ml_xSetWindowAttributes_set_backing_planes"
external winAttr_set_backing_pixel: xSetWindowAttributes -> backing_pixel:uint -> unit = "ml_xSetWindowAttributes_set_backing_pixel"
external winAttr_set_save_under: xSetWindowAttributes -> save_under:bool -> unit = "ml_xSetWindowAttributes_set_save_under"
external winAttr_set_event_mask: xSetWindowAttributes -> event_mask:event_mask_list -> unit = "ml_xSetWindowAttributes_set_event_mask"
external winAttr_set_do_not_propagate_mask: xSetWindowAttributes -> do_not_propagate_mask:int -> unit = "ml_xSetWindowAttributes_set_do_not_propagate_mask"
external winAttr_set_override_redirect: xSetWindowAttributes -> override_redirect:bool -> unit = "ml_xSetWindowAttributes_set_override_redirect"
external winAttr_set_colormap: xSetWindowAttributes -> colormap:colormap -> unit = "ml_xSetWindowAttributes_set_colormap"
external winAttr_set_cursor: xSetWindowAttributes -> cursor:cursor -> unit = "ml_xSetWindowAttributes_set_cursor"

(* }}} *)


(** {3 Windows} *)

external xRootWindow: dpy:display -> screen_number -> window = "ml_XRootWindow"
external xDefaultRootWindow: dpy:display -> window = "ml_XDefaultRootWindow"

external xCreateSimpleWindow: dpy:display -> parent:window -> x:int -> y:int ->
                              width:uint -> height:uint -> border_width:uint ->
                              border:pixel_color -> background:pixel_color -> window
    = "ml_XCreateSimpleWindow_bytecode"
      "ml_XCreateSimpleWindow"

external xDestroyWindow: dpy:display -> win:window -> unit = "ml_XDestroyWindow"

external xStoreName: dpy:display -> win:window -> name:string -> unit = "ml_XStoreName"

external xSelectInput: dpy:display -> win:window -> event_mask:event_mask list -> unit = "ml_XSelectInput"
external xMapWindow: dpy:display -> win:window -> unit = "ml_XMapWindow"


type window_class =
  | CopyFromParent
  | InputOutput
  | InputOnly

type winattr_valuemask =
  | CWBackPixmap
  | CWBackPixel
  | CWBorderPixmap
  | CWBorderPixel
  | CWBitGravity
  | CWWinGravity
  | CWBackingStore
  | CWBackingPlanes
  | CWBackingPixel
  | CWOverrideRedirect
  | CWSaveUnder
  | CWEventMask
  | CWDontPropagate
  | CWColormap
  | CWCursor

external xCreateWindow:
    dpy:display -> parent:window -> x:int -> y:int ->
    width:uint -> height:uint ->
    border_width:uint ->
    depth:int -> win_class:window_class -> visual:visual ->
    valuemask:winattr_valuemask list ->
    attributes:xSetWindowAttributes -> window
    = "ml_XCreateWindow_bytecode" "ml_XCreateWindow"


external xResizeWindow: dpy:display -> win:window -> width:uint -> height:uint -> unit = "ml_XResizeWindow"
external xLowerWindow: dpy:display -> win:window -> unit = "ml_XLowerWindow"
external xRaiseWindow: dpy:display -> win:window -> unit = "ml_XRaiseWindow"


(** {3 XScreen} *)

type xScreen
(** this type wraps the scruture Screen from the Xlib, but is here renamed with
    an additionnal x as a way of disambiguation with the type [screen_number] *)

external xDefaultScreenOfDisplay: dpy:display -> xScreen = "ml_XDefaultScreenOfDisplay"
(** the result points to a member of the input structure *)

external xScreenOfDisplay: dpy:display -> screen_number -> xScreen = "ml_XScreenOfDisplay"

external xDefaultVisualOfScreen: xScreen -> visual = "ml_XDefaultVisualOfScreen"
(** the result points to a member of the input structure *)

external xRootWindowOfScreen: xScreen -> window = "ml_XRootWindowOfScreen"
external xBlackPixelOfScreen: xScreen -> pixel_color = "ml_XBlackPixelOfScreen"
external xWhitePixelOfScreen: xScreen -> pixel_color = "ml_XWhitePixelOfScreen"
external xDefaultColormapOfScreen: xScreen -> colormap = "ml_XDefaultColormapOfScreen"
external xDefaultDepthOfScreen: xScreen -> int = "ml_XDefaultDepthOfScreen"
external xDefaultGCOfScreen: xScreen -> gc = "ml_XDefaultGCOfScreen"
external xWidthOfScreen: xScreen -> int = "ml_XWidthOfScreen"
external xHeightOfScreen: xScreen -> int = "ml_XHeightOfScreen"


(** {3 XSizeHints} *)

type xSizeHints
external new_xSizeHints: unit -> xSizeHints = "ml_alloc_XSizeHints"
(** this type is garbage collected *)

external xSizeHints_set_USPosition: xSizeHints -> x:int -> y:int -> unit = "ml_XSizeHints_set_USPosition"
external xSizeHints_set_USSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_USSize"
external xSizeHints_set_PPosition: xSizeHints -> x:int -> y:int -> unit = "ml_XSizeHints_set_PPosition"
external xSizeHints_set_PSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PSize"
external xSizeHints_set_PMinSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PMinSize"
external xSizeHints_set_PMaxSize: xSizeHints -> width:int -> height:int -> unit = "ml_XSizeHints_set_PMaxSize"
external xSizeHints_set_PResizeInc: xSizeHints -> width_inc:int -> height_inc:int -> unit = "ml_XSizeHints_set_PResizeInc"
external xSizeHints_set_PBaseSize: xSizeHints -> base_width:int -> base_height:int -> unit = "ml_XSizeHints_set_PBaseSize"
external xSizeHints_set_PAspect: xSizeHints -> min_aspect:int * int -> max_aspect:int * int -> unit = "ml_XSizeHints_set_PAspect"
external xSizeHints_set_PWinGravity: xSizeHints -> win_gravity:int -> unit = "ml_XSizeHints_set_PWinGravity"

external xSetNormalHints: dpy:display -> win:window -> hints:xSizeHints -> unit = "ml_XSetNormalHints"
external xSetStandardProperties: dpy:display -> win:window -> window_name:string -> icon_name:string ->
              icon_pixmap:pixmap option -> argv:string array -> hints:xSizeHints -> unit
    = "ml_XSetStandardProperties_bytecode"
      "ml_XSetStandardProperties"


(** {3 XEvents} *)

type 'a xEvent

type event_type =
  | KeyPress
  | KeyRelease
  | ButtonPress
  | ButtonRelease
  | MotionNotify
  | EnterNotify
  | LeaveNotify
  | FocusIn
  | FocusOut
  | KeymapNotify
  | Expose
  | GraphicsExpose
  | NoExpose
  | VisibilityNotify
  | CreateNotify
  | DestroyNotify
  | UnmapNotify
  | MapNotify
  | MapRequest
  | ReparentNotify
  | ConfigureNotify
  | ConfigureRequest
  | GravityNotify
  | ResizeRequest
  | CirculateNotify
  | CirculateRequest
  | PropertyNotify
  | SelectionClear
  | SelectionRequest
  | SelectionNotify
  | ColormapNotify
  | ClientMessage
  | MappingNotify


type xKeyEvent
type xButtonEvent
type xMotionEvent
type xCrossingEvent
type xFocusChangeEvent
type xExposeEvent
type xGraphicsExposeEvent
type xNoExposeEvent
type xVisibilityEvent
type xCreateWindowEvent
type xDestroyWindowEvent
type xUnmapEvent
type xMapEvent
type xMapRequestEvent
type xReparentEvent
type xConfigureEvent
type xGravityEvent
type xResizeRequestEvent
type xConfigureRequestEvent
type xCirculateEvent
type xCirculateRequestEvent
type xPropertyEvent
type xSelectionClearEvent
type xSelectionRequestEvent
type xSelectionEvent
type xColormapEvent
type xClientMessageEvent
type xMappingEvent
type xErrorEvent
type xKeymapEvent

type xKeyPressedEvent  = xKeyEvent
type xKeyReleasedEvent = xKeyEvent

type xButtonPressedEvent  = xButtonEvent
type xButtonReleasedEvent = xButtonEvent

type xEnterWindowEvent = xCrossingEvent
type xLeaveWindowEvent = xCrossingEvent

type xFocusInEvent  = xFocusChangeEvent
type xFocusOutEvent = xFocusChangeEvent


type event_kind =
  | XKeyEvent              of xKeyEvent              xEvent
  | XButtonEvent           of xButtonEvent           xEvent
  | XMotionEvent           of xMotionEvent           xEvent
  | XCrossingEvent         of xCrossingEvent         xEvent
  | XFocusChangeEvent      of xFocusChangeEvent      xEvent
  | XExposeEvent           of xExposeEvent           xEvent
  | XGraphicsExposeEvent   of xGraphicsExposeEvent   xEvent
  | XNoExposeEvent         of xNoExposeEvent         xEvent
  | XVisibilityEvent       of xVisibilityEvent       xEvent
  | XCreateWindowEvent     of xCreateWindowEvent     xEvent
  | XDestroyWindowEvent    of xDestroyWindowEvent    xEvent
  | XUnmapEvent            of xUnmapEvent            xEvent
  | XMapEvent              of xMapEvent              xEvent
  | XMapRequestEvent       of xMapRequestEvent       xEvent
  | XReparentEvent         of xReparentEvent         xEvent
  | XConfigureEvent        of xConfigureEvent        xEvent
  | XGravityEvent          of xGravityEvent          xEvent
  | XResizeRequestEvent    of xResizeRequestEvent    xEvent
  | XConfigureRequestEvent of xConfigureRequestEvent xEvent
  | XCirculateEvent        of xCirculateEvent        xEvent
  | XCirculateRequestEvent of xCirculateRequestEvent xEvent
  | XPropertyEvent         of xPropertyEvent         xEvent
  | XSelectionClearEvent   of xSelectionClearEvent   xEvent
  | XSelectionRequestEvent of xSelectionRequestEvent xEvent
  | XSelectionEvent        of xSelectionEvent        xEvent
  | XColormapEvent         of xColormapEvent         xEvent
  | XClientMessageEvent    of xClientMessageEvent    xEvent
  | XMappingEvent          of xMappingEvent          xEvent
  | XErrorEvent            of xErrorEvent            xEvent
  | XKeymapEvent           of xKeymapEvent           xEvent


external new_xEvent: unit -> 'a xEvent = "ml_alloc_XEvent"
(** the return value is garbage collected *)

external xNextEvent: dpy:display -> event:'a xEvent -> unit = "ml_XNextEvent"
external xNextEventFun: dpy:display -> 'a xEvent = "ml_XNextEvent_fun"
external xWindowEvent: dpy:display -> win:window -> event_mask list -> 'a xEvent = "ml_XWindowEvent"
external xPending: dpy:display -> int = "ml_XPending"

type queued_mode =
  | QueuedAlready
  | QueuedAfterFlush
  | QueuedAfterReading

external xEventsQueued: dpy:display -> mode:queued_mode -> int = "ml_XEventsQueued"

external xCheckTypedEvent: dpy:display -> event_type -> 'a xEvent -> bool = "ml_XCheckTypedEvent"

external xEventType: event:'a xEvent -> event_type = "ml_XEvent_type"


(** selects the right type of the event *)
let xEventKind ~event =
  match xEventType ~event with
  | MotionNotify        -> XMotionEvent           (Obj.magic event : xMotionEvent           xEvent)
  | KeyPress            -> XKeyEvent              (Obj.magic event : xKeyPressedEvent       xEvent)
  | KeyRelease          -> XKeyEvent              (Obj.magic event : xKeyReleasedEvent      xEvent)
  (* should we make 2 differents constructors for these cases?
                           XKeyPressedEvent
                           XKeyReleasedEvent *)
  | ButtonPress         -> XButtonEvent           (Obj.magic event : xButtonPressedEvent    xEvent)
  | ButtonRelease       -> XButtonEvent           (Obj.magic event : xButtonReleasedEvent   xEvent)
  | EnterNotify         -> XCrossingEvent         (Obj.magic event : xEnterWindowEvent      xEvent)
  | LeaveNotify         -> XCrossingEvent         (Obj.magic event : xLeaveWindowEvent      xEvent)
  | FocusIn             -> XFocusChangeEvent      (Obj.magic event : xFocusInEvent          xEvent)
  | FocusOut            -> XFocusChangeEvent      (Obj.magic event : xFocusOutEvent         xEvent)
 
  | KeymapNotify        -> XKeymapEvent           (Obj.magic event : xKeymapEvent           xEvent)
  | Expose              -> XExposeEvent           (Obj.magic event : xExposeEvent           xEvent)
  | GraphicsExpose      -> XGraphicsExposeEvent   (Obj.magic event : xGraphicsExposeEvent   xEvent)
  | NoExpose            -> XNoExposeEvent         (Obj.magic event : xNoExposeEvent         xEvent)
  | VisibilityNotify    -> XVisibilityEvent       (Obj.magic event : xVisibilityEvent       xEvent)
  | CreateNotify        -> XCreateWindowEvent     (Obj.magic event : xCreateWindowEvent     xEvent)
  | DestroyNotify       -> XDestroyWindowEvent    (Obj.magic event : xDestroyWindowEvent    xEvent)
  | UnmapNotify         -> XUnmapEvent            (Obj.magic event : xUnmapEvent            xEvent)
  | MapNotify           -> XMapEvent              (Obj.magic event : xMapEvent              xEvent)
  | MapRequest          -> XMapRequestEvent       (Obj.magic event : xMapRequestEvent       xEvent)
  | ReparentNotify      -> XReparentEvent         (Obj.magic event : xReparentEvent         xEvent)
  | ConfigureNotify     -> XConfigureEvent        (Obj.magic event : xConfigureEvent        xEvent)
  | ConfigureRequest    -> XConfigureRequestEvent (Obj.magic event : xConfigureRequestEvent xEvent)
  | GravityNotify       -> XGravityEvent          (Obj.magic event : xGravityEvent          xEvent)
  | ResizeRequest       -> XResizeRequestEvent    (Obj.magic event : xResizeRequestEvent    xEvent)
  | CirculateNotify     -> XCirculateEvent        (Obj.magic event : xCirculateEvent        xEvent)
  | CirculateRequest    -> XCirculateRequestEvent (Obj.magic event : xCirculateRequestEvent xEvent)
  | PropertyNotify      -> XPropertyEvent         (Obj.magic event : xPropertyEvent         xEvent)
  | SelectionClear      -> XSelectionClearEvent   (Obj.magic event : xSelectionClearEvent   xEvent)
  | SelectionRequest    -> XSelectionRequestEvent (Obj.magic event : xSelectionRequestEvent xEvent)
  | SelectionNotify     -> XSelectionEvent        (Obj.magic event : xSelectionEvent        xEvent)
  | ColormapNotify      -> XColormapEvent         (Obj.magic event : xColormapEvent         xEvent)
  | ClientMessage       -> XClientMessageEvent    (Obj.magic event : xClientMessageEvent    xEvent)
  | MappingNotify       -> XMappingEvent          (Obj.magic event : xMappingEvent          xEvent)
;;

(* {{{ XEvent conversion *)
(** {4 XEvent Conversions} *)

let to_xMotionEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MotionNotify then
    invalid_arg "to_xMotionEvent";
  (Obj.magic event : xMotionEvent  xEvent)
;;

let to_xKeyEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeyPress &&
     (xEventType ~event) <> KeyRelease then
    invalid_arg "to_xKeyEvent";
  (Obj.magic event : xKeyEvent  xEvent)
;;

let to_xKeyPressedEvent ~(event : 'a xEvent) =
  let t = (xEventType ~event) in
  if (xEventType ~event) <> KeyPress then
    begin
      Printf.eprintf "Event Type: %d\n%!" (Obj.magic t : int);
      if t = Expose then
        Printf.printf "10 confirmed\n%!";
      invalid_arg "to_xKeyPressedEvent";
    end;
  (Obj.magic event : xKeyPressedEvent  xEvent)
;;

let to_xKeyReleasedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeyRelease then
    invalid_arg "to_xKeyReleasedEvent";
  (Obj.magic event : xKeyReleasedEvent  xEvent)
;;

let to_xButtonEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonPress &&
     (xEventType ~event) <> ButtonRelease then
    invalid_arg "to_xButtonEvent";
  (Obj.magic event : xButtonEvent  xEvent)
;;

let to_xButtonPressedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonPress then
    invalid_arg "to_xButtonPressedEvent";
  (Obj.magic event : xButtonPressedEvent  xEvent)
;;

let to_xButtonReleasedEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ButtonRelease then
    invalid_arg "to_xButtonReleasedEvent";
  (Obj.magic event : xButtonReleasedEvent  xEvent)
;;

let to_xCrossingEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> EnterNotify &&
     (xEventType ~event) <> LeaveNotify then
    invalid_arg "to_xCrossingEvent";
  (Obj.magic event : xCrossingEvent  xEvent)
;;

let to_xEnterWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> EnterNotify then
    invalid_arg "to_xEnterWindowEvent";
  (Obj.magic event : xEnterWindowEvent  xEvent)
;;

let to_xLeaveWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> LeaveNotify then
    invalid_arg "to_xLeaveWindowEvent";
  (Obj.magic event : xLeaveWindowEvent  xEvent)
;;

let to_xFocusChangeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusIn &&
     (xEventType ~event) <> FocusOut then
    invalid_arg "to_xFocusChangeEvent";
  (Obj.magic event : xFocusChangeEvent  xEvent)
;;

let to_xFocusInEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusIn then
    invalid_arg "to_xFocusInEvent";
  (Obj.magic event : xFocusInEvent  xEvent)
;;

let to_xFocusOutEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> FocusOut then
    invalid_arg "to_xFocusOutEvent";
  (Obj.magic event : xFocusOutEvent  xEvent)
;;

let to_xKeymapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> KeymapNotify then
    invalid_arg "to_xKeymapEvent";
  (Obj.magic event : xKeymapEvent  xEvent)
;;

let to_xExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> Expose then
    invalid_arg "to_xExposeEvent";
  (Obj.magic event : xExposeEvent  xEvent)
;;

let to_xGraphicsExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> GraphicsExpose then
    invalid_arg "to_xGraphicsExposeEvent";
  (Obj.magic event : xGraphicsExposeEvent  xEvent)
;;

let to_xNoExposeEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> NoExpose then
    invalid_arg "to_xNoExposeEvent";
  (Obj.magic event : xNoExposeEvent  xEvent)
;;

let to_xVisibilityEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> VisibilityNotify then
    invalid_arg "to_xVisibilityEvent";
  (Obj.magic event : xVisibilityEvent  xEvent)
;;

let to_xCreateWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CreateNotify then
    invalid_arg "to_xCreateWindowEvent";
  (Obj.magic event : xCreateWindowEvent  xEvent)
;;

let to_xDestroyWindowEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> DestroyNotify then
    invalid_arg "to_xDestroyWindowEvent";
  (Obj.magic event : xDestroyWindowEvent  xEvent)
;;

let to_xUnmapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> UnmapNotify then
    invalid_arg "to_xUnmapEvent";
  (Obj.magic event : xUnmapEvent  xEvent)
;;

let to_xMapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MapNotify then
    invalid_arg "to_xMapEvent";
  (Obj.magic event : xMapEvent  xEvent)
;;

let to_xMapRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MapRequest then
    invalid_arg "to_xMapRequestEvent";
  (Obj.magic event : xMapRequestEvent  xEvent)
;;

let to_xReparentEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ReparentNotify then
    invalid_arg "to_xReparentEvent";
  (Obj.magic event : xReparentEvent  xEvent)
;;

let to_xConfigureEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ConfigureNotify then
    invalid_arg "to_xConfigureEvent";
  (Obj.magic event : xConfigureEvent  xEvent)
;;

let to_xConfigureRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ConfigureRequest then
    invalid_arg "to_xConfigureRequestEvent";
  (Obj.magic event : xConfigureRequestEvent  xEvent)
;;

let to_xGravityEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> GravityNotify then
    invalid_arg "to_xGravityEvent";
  (Obj.magic event : xGravityEvent  xEvent)
;;

let to_xResizeRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ResizeRequest then
    invalid_arg "to_xResizeRequestEvent";
  (Obj.magic event : xResizeRequestEvent  xEvent)
;;

let to_xCirculateEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CirculateNotify then
    invalid_arg "to_xCirculateEvent";
  (Obj.magic event : xCirculateEvent  xEvent)
;;

let to_xCirculateRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> CirculateRequest then
    invalid_arg "to_xCirculateRequestEvent";
  (Obj.magic event : xCirculateRequestEvent  xEvent)
;;

let to_xPropertyEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> PropertyNotify then
    invalid_arg "to_xPropertyEvent";
  (Obj.magic event : xPropertyEvent  xEvent)
;;

let to_xSelectionClearEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionClear then
    invalid_arg "to_xSelectionClearEvent";
  (Obj.magic event : xSelectionClearEvent  xEvent)
;;

let to_xSelectionRequestEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionRequest then
    invalid_arg "to_xSelectionRequestEvent";
  (Obj.magic event : xSelectionRequestEvent  xEvent)
;;

let to_xSelectionEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> SelectionNotify then
    invalid_arg "to_xSelectionEvent";
  (Obj.magic event : xSelectionEvent  xEvent)
;;

let to_xColormapEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ColormapNotify then
    invalid_arg "to_xColormapEvent";
  (Obj.magic event : xColormapEvent  xEvent)
;;

let to_xClientMessageEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> ClientMessage then
    invalid_arg "to_xClientMessageEvent";
  (Obj.magic event : xClientMessageEvent  xEvent)
;;

let to_xMappingEvent ~(event : 'a xEvent) =
  if (xEventType ~event) <> MappingNotify then
    invalid_arg "to_xMappingEvent";
  (Obj.magic event : xMappingEvent  xEvent)
;;

(* }}} *)
(* {{{ XEvents contents *)
(** {4 XEvents Contents} *)

type xKeyEvent_contents = {
    key_time: int;
    key_x: int;
    key_y: int;
    key_x_root: int;
    key_y_root: int;
    key_state: uint;
    key_keycode: uint;
    key_same_screen: bool;
  }
external xKeyEvent_datas: xKeyEvent xEvent -> xKeyEvent_contents = "ml_xKeyEvent_datas"

type xConfigureEvent_contents = {
    conf_x: int;
    conf_y: int;
    conf_width: int;
    conf_height: int;
    conf_border_width: int;
    conf_above: window;
    conf_override_redirect: bool;
  }
external xConfigureEvent_datas: xConfigureEvent xEvent -> xConfigureEvent_contents = "ml_xConfigureEvent_datas"

type xMotionEvent_contents = {
    motion_root: window;
    motion_subwindow: window;
    motion_time: int;
    motion_x: int;
    motion_y: int;
    motion_x_root: int;
    motion_y_root: int;
    motion_state: uint;
    motion_is_hint: char;
    motion_same_screen: bool;
  }
external xMotionEvent_datas: xMotionEvent xEvent -> xMotionEvent_contents = "ml_xMotionEvent_datas"

type button =
  | AnyButton
  | Button1
  | Button2
  | Button3
  | Button4
  | Button5

type xButtonEvent_contents = {
    button_root: window;
    button_subwindow: window;
    button_time: int;
    button_x: int;
    button_y: int;
    button_x_root: int;
    button_y_root: int;
    button_state: uint;
    button: button;
    button_same_screen: bool;
  }
external xButtonEvent_datas: xButtonEvent xEvent -> xButtonEvent_contents = "ml_xButtonEvent_datas"

(* }}} *)

type keysym = int
external xLookupString: event:xKeyEvent xEvent -> buffer:string -> int * keysym = "ml_XLookupString"

external xLookupKeysym: event:xKeyEvent xEvent -> index:int -> keysym = "ml_XLookupKeysym"
external xKeycodeToKeysym: dpy:display -> keycode:uint -> index:int -> keysym = "ml_XKeycodeToKeysym"

external xRefreshKeyboardMapping : event:xMappingEvent xEvent -> unit = "ml_XRefreshKeyboardMapping"


(** {3 Atoms} *)

type atom
external xSetWMProtocols: dpy:display -> win:window -> protocols:atom -> count:int -> unit = "ml_XSetWMProtocols"
external xInternAtom: dpy:display -> atom_name:string -> only_if_exists:bool -> atom = "ml_XInternAtom"

external xEvent_xclient_data : xClientMessageEvent xEvent -> atom = "ml_xEvent_xclient_data_l_0"


(** {3 GC (Graphics Context)} *)

external xDefaultGC: dpy:display -> screen_number -> gc = "ml_XDefaultGC"

type xGCValues
external new_xGCValues: unit -> xGCValues = "ml_alloc_XGCValues"
(** the return value is garbage collected *)

external xGCValues_set_foreground: gcv:xGCValues -> fg:pixel_color -> unit = "ml_XGCValues_set_foreground"
external xGCValues_set_background: gcv:xGCValues -> bg:pixel_color -> unit = "ml_XGCValues_set_background"

type gc_valuemask =
  | GCFunction
  | GCPlaneMask
  | GCForeground
  | GCBackground
  | GCLineWidth
  | GCLineStyle
  | GCCapStyle
  | GCJoinStyle
  | GCFillStyle
  | GCFillRule
  | GCTile
  | GCStipple
  | GCTileStipXOrigin
  | GCTileStipYOrigin
  | GCFont
  | GCSubwindowMode
  | GCGraphicsExposures
  | GCClipXOrigin
  | GCClipYOrigin
  | GCClipMask
  | GCDashOffset
  | GCDashList
  | GCArcMode

type drawable
external xCreateGC: dpy:display -> d:drawable -> gc_valuemask list -> xGCValues -> gc = "ml_XCreateGC"
external xFreeGC: dpy:display -> gc:gc -> unit = "ml_XFreeGC"


(** {3 Drawing} *)

let drawable_of_window ~(win : window) = (Obj.magic win : drawable)
let drawable_of_pixmap ~(win : pixmap) = (Obj.magic win : drawable)

external xSetForeground: dpy:display -> gc:gc -> fg:pixel_color -> unit = "ml_XSetForeground"
external xSetBackground: dpy:display -> gc:gc -> bg:pixel_color -> unit = "ml_XSetBackground"

external xClearWindow: dpy:display -> win:window -> unit = "ml_XClearWindow"

external xDrawString: dpy:display -> d:drawable -> gc:gc ->
                      x:int -> y:int -> str:string -> unit
    = "ml_XDrawString_bytecode"
      "ml_XDrawString"

external xFillArc: dpy:display -> d:drawable -> gc:gc -> x:int -> y:int ->
                   width:uint -> height:uint -> angle1:int -> angle2:int -> unit
    = "ml_XFillArc_bytecode"
      "ml_XFillArc"

external xDrawArc: dpy:display -> d:drawable -> gc:gc -> x:int -> y:int ->
                   width:uint -> height:uint -> angle1:int -> angle2:int -> unit
    = "ml_XDrawArc_bytecode"
      "ml_XDrawArc"

external xDrawRectangle: dpy:display -> d:drawable -> gc:gc ->
                         x:int -> y:int -> width:uint -> height:uint -> unit
    = "ml_XDrawRectangle_bytecode"
      "ml_XDrawRectangle"

type segment = {
    x1:int;
    y1:int;
    x2:int;
    y2:int;
  }
external xDrawSegments: dpy:display -> d:drawable -> gc:gc -> segments:segment array -> unit = "ml_XDrawSegments"

external xFillRectangle: dpy:display -> d:drawable -> gc:gc ->
                         x:int -> y:int -> width:uint -> height:uint -> unit
    = "ml_XFillRectangle_bytecode"
      "ml_XFillRectangle"


(** {3 Font} *)

type font
external xLoadFont: dpy:display -> name:string -> font = "ml_XLoadFont"


(** {3 Dbl} *)

external xCopyArea: dpy:display ->
    src:drawable ->
    dest:drawable ->
    gc:gc ->
    src_x:int ->
    src_y:int ->
    width:uint ->
    height: uint ->
    dest_x:int ->
    dest_y:int -> unit
    = "ml_XCopyArea_bytecode"
      "ml_XCopyArea"

external xCreatePixmap: dpy:display -> d:drawable -> width:uint -> height:uint -> depth:uint -> pixmap = "ml_XCreatePixmap"
external xFreePixmap: dpy:display -> pixmap -> unit = "ml_XFreePixmap"


(** {4 Keyboard} *)

external xAutoRepeatOff: dpy:display -> unit = "ml_XAutoRepeatOff"
external xAutoRepeatOn: dpy:display -> unit = "ml_XAutoRepeatOn"
external xQueryKeymap: dpy:display -> string = "ml_XQueryKeymap"

module Got = struct
type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn
end
module Set = struct
type auto_repeat_mode = AutoRepeatModeOff | AutoRepeatModeOn | AutoRepeatModeDefault
end
type xKeyboardState = {
    key_click_percent: int;
    bell_percent: int;
    bell_pitch: uint;
    bell_duration: uint;
    led_mask: uint;
    global_auto_repeat: Got.auto_repeat_mode;
    auto_repeats: string;
  }
external xGetKeyboardControl: dpy:display -> xKeyboardState = "ml_XGetKeyboardControl"


(** {4 screensaver} *)

type screensaver_mode =
  | ScreenSaverActive
  | ScreenSaverReset

external xForceScreenSaver: dpy:display -> mode:screensaver_mode -> unit = "ml_XForceScreenSaver"

type screensaver_values = {
    timeout: int;
    interval: int;
    prefer_blanking: int;
    allow_exposures: int;
  }
external xGetScreenSaver: dpy:display -> screensaver_values = "ml_XGetScreenSaver"


(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
